@@ -3,14 +3,19 @@
 
 #include "device.h"
 
+#ifdef _WIN32
+#include <windows.h>
+#endif
+
 #include <algorithm>
+#include <bit>
+#include <csignal>
 #include <cstddef>
 #include <numeric>
 
 #include "al/buffer.h"
 #include "al/effect.h"
 #include "al/filter.h"
-#include "albit.h"
 #include "alnumeric.h"
 #include "atomic.h"
 #include "backends/base.h"
@@ -19,6 +24,7 @@
 #include "core/logging.h"
 #include "core/mastering.h"
 #include "flexarray.h"
+#include "gsl/gsl"
 
 
 namespace {
@@ -27,49 +33,55 @@ using voidp = void*;
 
 } // namespace
 
+namespace al {
 
-ALCdevice::ALCdevice(DeviceType type) : DeviceBase{type}
+void DeviceDeleter::operator()(gsl::owner<Device*> device) const noexcept
+{ delete device; }
+
+auto Device::Create(DeviceType type) -> intrusive_ptr<Device>
+{ return intrusive_ptr{new Device{type}}; }
+
+Device::Device(DeviceType type) : DeviceBase{type}
 { }
 
-ALCdevice::~ALCdevice()
+Device::~Device()
 {
-    TRACE("Freeing device %p\n", voidp{this});
+    TRACE("Freeing device {}", voidp{this});
 
     Backend = nullptr;
 
-    size_t count{std::accumulate(BufferList.cbegin(), BufferList.cend(), 0_uz,
+    auto count = std::accumulate(BufferList.cbegin(), BufferList.cend(), 0_uz,
         [](size_t cur, const BufferSubList &sublist) noexcept -> size_t
-        { return cur + static_cast<uint>(al::popcount(~sublist.FreeMask)); })};
+        { return cur + gsl::narrow_cast<uint>(std::popcount(~sublist.mFreeMask)); });
     if(count > 0)
-        WARN("%zu Buffer%s not deleted\n", count, (count==1)?"":"s");
+        WARN("{} Buffer{} not deleted", count, (count==1)?"":"s");
 
     count = std::accumulate(EffectList.cbegin(), EffectList.cend(), 0_uz,
         [](size_t cur, const EffectSubList &sublist) noexcept -> size_t
-        { return cur + static_cast<uint>(al::popcount(~sublist.FreeMask)); });
+        { return cur + gsl::narrow_cast<uint>(std::popcount(~sublist.mFreeMask)); });
     if(count > 0)
-        WARN("%zu Effect%s not deleted\n", count, (count==1)?"":"s");
+        WARN("{} Effect{} not deleted", count, (count==1)?"":"s");
 
     count = std::accumulate(FilterList.cbegin(), FilterList.cend(), 0_uz,
         [](size_t cur, const FilterSubList &sublist) noexcept -> size_t
-        { return cur + static_cast<uint>(al::popcount(~sublist.FreeMask)); });
+        { return cur + gsl::narrow_cast<uint>(std::popcount(~sublist.mFreeMask)); });
     if(count > 0)
-        WARN("%zu Filter%s not deleted\n", count, (count==1)?"":"s");
+        WARN("{} Filter{} not deleted", count, (count==1)?"":"s");
 }
 
-void ALCdevice::enumerateHrtfs()
+void Device::enumerateHrtfs()
 {
     mHrtfList = EnumerateHrtf(configValue<std::string>({}, "hrtf-paths"));
     if(auto defhrtfopt = configValue<std::string>({}, "default-hrtf"))
     {
-        auto iter = std::find(mHrtfList.begin(), mHrtfList.end(), *defhrtfopt);
-        if(iter == mHrtfList.end())
-            WARN("Failed to find default HRTF \"%s\"\n", defhrtfopt->c_str());
+        if(const auto iter = std::ranges::find(mHrtfList, *defhrtfopt); iter == mHrtfList.end())
+            WARN("Failed to find default HRTF \"{}\"", *defhrtfopt);
         else if(iter != mHrtfList.begin())
             std::rotate(mHrtfList.begin(), iter, iter+1);
     }
 }
 
-auto ALCdevice::getOutputMode1() const noexcept -> OutputMode1
+auto Device::getOutputMode1() const noexcept -> OutputMode1
 {
     if(mContexts.load(std::memory_order_relaxed)->empty())
         return OutputMode1::Any;
@@ -80,7 +92,7 @@ auto ALCdevice::getOutputMode1() const noexcept -> OutputMode1
     case DevFmtStereo:
         if(mHrtf)
             return OutputMode1::Hrtf;
-        else if(mUhjEncoder)
+        if(std::holds_alternative<UhjPostProcess>(mPostProcess))
             return OutputMode1::Uhj2;
         return OutputMode1::StereoBasic;
     case DevFmtQuad: return OutputMode1::Quad;
@@ -95,3 +107,26 @@ auto ALCdevice::getOutputMode1() const noexcept -> OutputMode1
     }
     return OutputMode1::Any;
 }
+
+
+void Device::SetError(Device *device, ALCenum errorCode)
+{
+    WARN("Error generated on device {}, code {:#04x}", voidp{device}, as_unsigned(errorCode));
+    if(sTrapALCError)
+    {
+#ifdef _WIN32
+        /* DebugBreak() will cause an exception if there is no debugger */
+        if(IsDebuggerPresent())
+            DebugBreak();
+#elif defined(SIGTRAP)
+        raise(SIGTRAP);
+#endif
+    }
+
+    if(device)
+        device->mLastError.store(errorCode);
+    else
+        sLastGlobalError.store(errorCode);
+}
+
+} // namespace al
